import pytest

from game_theory import NegativeNumberException
from game_theory.algorithms.bankruptcy import (
    constrained_equal_awards,
    constrained_equal_losses,
)


def test_constrained_equal_awards_two_person_equal():
    claims = [50, 60]
    asset = 50
    allocations, r = constrained_equal_awards(claims, asset)

    assert allocations == [25, 25]
    assert r == 25


def test_constrained_equal_awards_two_person_unequal():
    claims = [50, 24]
    asset = 50
    allocations, r = constrained_equal_awards(claims, asset)

    assert allocations == [26, 24]
    assert r == 26


def test_constrained_equal_awards_three_person_equal():
    claims = [50, 24, 68]
    asset = 60
    allocations, r = constrained_equal_awards(claims, asset)

    assert allocations == [20, 20, 20]
    assert r == 20


def test_constrained_equal_awards_three_person_unequal():
    claims = [50, 24, 68]
    asset = 75
    allocations, r = constrained_equal_awards(claims, asset)

    assert allocations == [25.5, 24, 25.5]
    assert r == 25.5


def test_constrained_equal_awards_same_allocation():
    claims = [50, 24, 68, 80]
    asset = 1000
    allocations, r = constrained_equal_awards(claims, asset)

    assert allocations == [50, 24, 68, 80]
    assert r == 80


def test_constrained_equal_awards_non_negative_asset():
    with pytest.raises(
        NegativeNumberException, match="Asset cannot be a negative number."
    ):
        claims = [50, 24, 68]
        asset = -25
        constrained_equal_awards(claims, asset)


def test_constrained_equal_awards_non_negative_claims():
    with pytest.raises(
        NegativeNumberException, match="Claims cannot be a negative number."
    ):
        claims = [50, -24, 68]
        asset = 25
        constrained_equal_awards(claims, asset)


def test_constrained_equal_losses_two_person_equal_loss_1():
    claims = [60, 90]
    asset = 100
    allocations, r = constrained_equal_losses(claims, asset)
    assert allocations == [35, 65]
    assert r == 25


def test_constrained_equal_losses_two_person_equal_loss_2():
    claims = [50, 100]
    asset = 100
    allocations, r = constrained_equal_losses(claims, asset)
    assert allocations == [25, 75]
    assert r == 25


def test_constrained_equal_losses_two_person_equal_loss_3():
    claims = [40, 80]
    asset = 100
    allocations, r = constrained_equal_losses(claims, asset)
    assert allocations == [30, 70]
    assert r == 10


def test_constrained_equal_losses_two_person_unequal_loss():
    claims = [450, 50]
    asset = 100
    allocations, r = constrained_equal_losses(claims, asset)
    assert allocations == [100, 0]
    assert r == 350


def test_constrained_equal_losses_three_person_unequal_loss():
    claims = [50, 100, 150]
    asset = 100
    allocations, r = constrained_equal_losses(claims, asset)
    assert allocations == [0, 25, 75]
    assert r == 75


def test_constrained_equal_losses_three_person_same_allocation():
    claims = [50, 100, 150]
    asset = 300
    allocations, r = constrained_equal_losses(claims, asset)
    assert allocations == [50, 100, 150]
    assert r == 0


def test_constrained_equal_losses_non_negative_asset():
    with pytest.raises(
        NegativeNumberException, match="Asset cannot be a negative number."
    ):
        claims = [50, 24, 68]
        asset = -25
        constrained_equal_losses(claims, asset)


def test_constrained_equal_losses_non_negative_claims():
    with pytest.raises(
        NegativeNumberException, match="Claims cannot be a negative number."
    ):
        claims = [50, -24, 68]
        asset = 25
        constrained_equal_losses(claims, asset)



"""
Module for implementing classical auction mechanisms from game theory.
"""

from game_theory import NegativeNumberException


def first_price_auction(bids):
    """
    First-Price Sealed-Bid Auction:
    Highest bidder wins and pays their bid.

    Args:
        bids (list): A list of non-negative bids (floats or ints).

    Returns:
        winner (int): Index of the winning bidder.
        price (float): Price paid (equal to winning bid).
    """
    if not bids:
        raise ValueError("Bids list cannot be empty.")

    if any(b < 0 for b in bids):
        raise NegativeNumberException("Bids must be non-negative.", bids)

    winner = max(range(len(bids)), key=lambda i: bids[i])
    return winner, bids[winner]


def second_price_auction(bids):
    """
    Second-Price (Vickrey) Sealed-Bid Auction:
    Highest bidder wins and pays the second-highest bid.

    If the winner is not unique, the winner is the last bidder index with the highest bid.

    Args:
        bids (list): A list of non-negative bids.

    Returns:
        winner (int): Index of winning bidder.
        price (float): Price paid (second-highest bid).
    """
    if not bids:
        raise ValueError("Bids list cannot be empty.")

    if any(b < 0 for b in bids):
        raise NegativeNumberException("Bids must be non-negative.", bids)

    sorted_bids = sorted(((bid, i) for i, bid in enumerate(bids)), reverse=True)
    highest_bid, winner = sorted_bids[0]
    second_highest_bid = sorted_bids[1][0] if len(bids) > 1 else 0.0

    return winner, second_highest_bid


def generalized_second_price_auction(bids, slots):
    """
    Generalized Second-Price Auction (e.g., search ads).

    Args:
        bids (list): A list of non-negative bids (CPC).
        slots (list): A list of slot click-through rates (CTR), descending order.

    Returns:
        allocation (list of tuples): (bidder index, assigned slot, price per click).
    """
    if not bids or not slots:
        raise ValueError("Bids and slots must be non-empty lists.")

    if any(b < 0 for b in bids):
        raise NegativeNumberException("Bids must be non-negative.", bids)

    if any(s < 0 for s in slots):
        raise NegativeNumberException("Slots must be non-negative.", slots)

    if len(bids) < len(slots):
        slots = slots[: len(bids)]

    sorted_bidders = sorted(((bid, i) for i, bid in enumerate(bids)), reverse=True)
    allocation = []

    for j, (slot_ctr) in enumerate(slots):
        if j >= len(sorted_bidders):
            break
        _, bidder = sorted_bidders[j]
        next_bid = sorted_bidders[j + 1][0] if j + 1 < len(sorted_bidders) else 0.0
        price = next_bid  # Pay the bid of the next highest bidder
        allocation.append((bidder, j, price))

    return allocation



from game_theory.algorithms.auction import (
    first_price_auction,
    generalized_second_price_auction,
    second_price_auction,
)
from game_theory.algorithms.bankruptcy import (
    CEA,
    CEL,
    constrained_equal_awards,
    constrained_equal_losses,
)



class NegativeNumberException(Exception):
    """Exception raised for negative numbers.
    """

    def __init__(self, message, value):
        self.message = message if message else "Negative numbers are not allowed."
        self.value = value
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message} -> {self.value}"



import pytest

from game_theory import NegativeNumberException
from game_theory.algorithms.auction import (
    first_price_auction,
    generalized_second_price_auction,
    second_price_auction,
)


# Tests for first_price_auction
def test_first_price_auction_positive():
    assert first_price_auction([1, 5, 3]) == (1, 5)


def test_first_price_auction_edge_all_equal():
    assert first_price_auction([2, 2, 2]) == (0, 2)


def test_first_price_auction_edge_single_bidder():
    assert first_price_auction([10]) == (0, 10)


def test_first_price_auction_error_empty():
    with pytest.raises(ValueError):
        first_price_auction([])


def test_first_price_auction_error_negative_bid():
    with pytest.raises(NegativeNumberException):
        first_price_auction([5, -2, 3])


# Tests for second_price_auction
def test_second_price_auction_positive():
    assert second_price_auction([4, 6, 2]) == (1, 4)


def test_second_price_auction_edge_all_equal():
    # the winner is the last bidder with the highest bid
    assert second_price_auction([3, 3, 3]) == (2, 3)


def test_second_price_auction_edge_single_bidder():
    assert second_price_auction([8]) == (0, 0)


def test_second_price_auction_error_empty():
    with pytest.raises(ValueError):
        second_price_auction([])


def test_second_price_auction_error_negative_bid():
    with pytest.raises(NegativeNumberException):
        second_price_auction([1, -3, 2])


# Tests for generalized_second_price_auction
def test_gspa_positive():
    assert generalized_second_price_auction([10, 9, 8], [0.5, 0.3]) == [
        (0, 0, 9),
        (1, 1, 8),
    ]


def test_gspa_edge_one_slot():
    assert generalized_second_price_auction([7], [0.4]) == [(0, 0, 0)]


def test_gspa_edge_extra_bidders():
    assert generalized_second_price_auction([9, 6, 4], [0.5]) == [(0, 0, 6)]


def test_gspa_error_empty_bids():
    with pytest.raises(ValueError):
        generalized_second_price_auction([], [0.5])


def test_gspa_error_negative_slot():
    with pytest.raises(NegativeNumberException):
        generalized_second_price_auction([5], [-0.1])




"""
Module for division rules for solving bankruptcy problems.
"""

from game_theory import NegativeNumberException


def constrained_equal_awards(claims, asset):
    """
    Calculate allocations using the constrained equal awards method.
    If the asset is larger than the sum of claims, returns the claims list itself.

    Args:
        claims (list): A list of claims. Each element of the list is a non-negative float number.
        asset (float): The total amount of available assets. This is a non-negative number.

    Raises:
        NegativeNumberException: Claims cannot be a negative number.

    Returns:
        list: A list of allocations.
        float: The r value in the equation a_i = min(r, c_i). If the sum of claims is less than the asset,
        then the r-value is the maximum claim value.
    """

    if any(c < 0 for c in claims):
        raise NegativeNumberException(
            "Claims cannot be a negative number.", claims)

    if asset < 0:
        raise NegativeNumberException(
            "Asset cannot be a negative number.", asset)

    # sum of claims is less than the asset (no bankruptcy)
    if sum(claims) <= asset:
        return claims, max(claims)

    equal_alloc = asset / len(claims)

    if all(c >= equal_alloc for c in claims):
        return [equal_alloc for _ in range(len(claims))], equal_alloc

    r = min(claims)
    allocations = [r for _ in range(len(claims))]

    while asset > sum(allocations):
        remained_claims = [c - r for c in claims]
        r_prime = min(c for c in remained_claims if c > 0)
        positive_remained_claim_length = len(
            [c for c in remained_claims if c > 0])

        if asset - (sum(allocations) + r_prime * positive_remained_claim_length) < 0:
            r_prime = (asset - sum(allocations)) / \
                positive_remained_claim_length

        allocations = [alloc + r_prime if remained > 0 else alloc for alloc,
                       remained in zip(allocations, remained_claims)]
        r += r_prime

    return allocations, r


def CEA(claims, asset):
    """
    Calculate allocations using the constrained equal awards method.

    Args:
        claims (list): A list of claims. Each element of the list is a non-negative float number.
        asset (float): The total amount of available assets. This is a non-negative number.

    Raises:
        NegativeNumberException: Claims cannot be a negative number.

    Returns:
        list: A list of allocations.
        float: The r value in the equation a_i = min(r, c_i). If the sum of claims is less than the asset,
        then the r-value is the maximum claim value.
    """
    return constrained_equal_awards(claims, asset)


def constrained_equal_losses(claims, asset):
    """
    Calculate allocations using the constrained equal losses method.
    If the asset is larger than the sum of claims, returns the claims list itself.

    Raises:
        NegativeNumberException: Claims cannot be a negative number.

    Returns:
        list: A list of allocations.
        float: The r value in the equation a_i = max(0, c_i-r). If the sum of claims is less than the asset,
        then the r-value is 0.
    """

    # checks
    if any(c < 0 for c in claims):
        raise NegativeNumberException(
            "Claims cannot be a negative number.", claims)

    if asset < 0:
        raise NegativeNumberException(
            "Asset cannot be a negative number.", asset)

    total_loss = sum(claims) - asset

    if total_loss <= 0:
        return claims, 0

    equal_loss = total_loss / len(claims)

    if all(c >= equal_loss for c in claims):
        return [c - equal_loss for c in claims], equal_loss

    r = min(claims)
    losses = [r for _ in range(len(claims))]

    while total_loss > sum(losses):
        remained_claims = [c - r for c in claims]
        r_prime = min(c for c in remained_claims if c > 0)
        positive_remained_claim_length = len(
            [c for c in remained_claims if c > 0])

        if total_loss - (sum(losses) + r_prime * positive_remained_claim_length) < 0:
            r_prime = (total_loss - sum(losses)) / \
                positive_remained_claim_length

        losses = [alloc + r_prime if remained > 0 else alloc for alloc,
                  remained in zip(losses, remained_claims)]
        r += r_prime

    return [claim - loss for claim, loss in zip(claims, losses)], r


def CEL(claims, asset):
    """
    Calculate allocations using the constrained equal losses method.
    If the asset is larger than the sum of claims, returns the claims list itself.

    Raises:
        NegativeNumberException: Claims cannot be a negative number.

    Returns:
        list: A list of allocations.
        float: The r value in the equation a_i = max(0, c_i-r). If the sum of claims is less than the asset,
        then the r-value is 0.
    """
    return constrained_equal_losses(claims, asset)